import pathlib
from typing import Dict

import openapi3.object_base
import requests
import yaml
from dict_tools.data import NamespaceDict


def context(hub, ctx, directory: pathlib.Path):
    ctx = hub.pop_create.idem_cloud.init.context(ctx, directory)

    spec = hub.pop_create.openapi3.init.read(source=ctx.specification)

    api = openapi3.OpenAPI(spec, validate=True)
    errors = api.errors()
    if errors:
        for e in errors:
            hub.log.warning(e)

    # list these as defaults in the acct plugin
    if api.servers:
        ctx.servers = [x.url for x in api.servers]
    else:
        ctx.servers = ["https://"]

    hub.log.debug(f"Working with openapi spec version: {api.openapi}")
    ctx.cloud_api_version = api.info.version or "latest"
    ctx.clean_api_version = hub.tool.format.case.snake(ctx.cloud_api_version).strip("_")
    # If the api version starts with a digit then make sure it can be used for python namespacing
    if ctx.clean_api_version[0].isdigit():
        ctx.clean_api_version = "v" + ctx.clean_api_version

    cloud_spec = NamespaceDict(
        api_version=ctx.cloud_api_version,
        project_name=ctx.project_name,
        service_name=ctx.service_name,
        request_format=hub.pop_create.openapi3.template.HTTP_REQUEST.format(
            acct_plugin=ctx.acct_plugin
        ),
        plugins=hub.pop_create.openapi3.parse.plugins(ctx, api.paths),
    )
    ctx.cloud_spec = cloud_spec

    hub.pop_create.init.run(directory=directory, subparsers=["idem_cloud"], **ctx)
    return ctx


def read(hub, source: str or Dict):
    """
    If the path is a file, then parse the json contents of the file,
    If the path is a url, then return a json response from the url.
    """
    if isinstance(source, Dict):
        return source

    path = pathlib.Path(source)

    if path.exists():
        with path.open("r") as fh:
            ret = yaml.safe_load(fh)
    else:
        request = requests.get(source, headers={"Content-Type": "application/json"})
        ret = request.json()

    return ret



import pathlib

from cloudspec import CloudSpec


def run(hub, ctx, root_directory: pathlib.Path or str):
    if isinstance(root_directory, str):
        root_directory = pathlib.Path(root_directory)
    cloud_spec = CloudSpec(**ctx.cloud_spec)
    exec_dir = root_directory / ctx.clean_name / "exec" / ctx.service_name

    for ref, plugin in cloud_spec.plugins.items():
        mod_file = hub.cloudspec.parse.plugin.touch(exec_dir, ref)
        ref = hub.cloudspec.parse.plugin.ref(ctx, ref)
        cli_ref = hub.cloudspec.parse.plugin.mod_ref(ctx, ref, plugin)

        to_write = hub.cloudspec.parse.plugin.header(plugin)

        for function_name, function_data in plugin.functions.items():
            template = hub.tool.jinja.template(
                f"{hub.cloudspec.template.exec.FUNCTION}\n    {cloud_spec.request_format}\n\n\n"
            )

            function_alias = plugin.func_alias.get(function_name, function_name)

            to_write += template.render(
                function=dict(
                    name=function_name,
                    hardcoded=function_data.hardcoded,
                    doc=hub.cloudspec.create.function.doc(function_data)
                    + hub.cloudspec.parse.param.sphinx_doc(function_data.params)
                    + hub.cloudspec.parse.function.return_type(function_data),
                    ref=f"{ref}.{function_alias}",
                    cli_ref=f"{cli_ref}.{function_alias}",
                    header_params=hub.cloudspec.parse.param.headers(
                        function_data.params
                    ),
                    required_call_params=hub.cloudspec.parse.param.callers(
                        function_data.params
                    ),
                    return_type=function_data.return_type,
                ),
                parameter=dict(
                    mapping=hub.cloudspec.parse.param.mappings_get_map_params(
                        function_data.params
                    )
                ),
            )

        mod_file.write_text(to_write)



import pathlib


EXEC_UNIT_TEST = r'''
@pytest.mark.asyncio
async def test_\{\{function.name\}\}(hub, mock_hub, ctx):
    """
    Autogenerated function

    Test whether a call is successful and uses the expected parameters
    """
    mock_hub.exec.{{cookiecutter.service_name}}.R__SUBS__.R__FUNC_NAME = (
        hub.exec.{{cookiecutter.service_name}}.R__SUBS__.R__FUNC_NAME
    )

    await hub.exec.{{cookiecutter.service_name}}.\{\{plugin.subs\}\}.\{\{function.name\}\}(ctx, \{\{function.required_kwargs\}\})
    hub.tool.{{cookiecutter.service_name}}.session.request.assert_called_once_with(ctx)
'''

EXEC_INTEGRATION_TEST = r"""
@pytest.mark.asyncio
async def test_\{\{function.name\}\}(hub, ctx):
    ret = await hub.exec.{{cookiecutter.service_name}}.\{\{plugin.subs\}\}.\{\{plugin.name\}\}.\{\{function.name\}\}(
        ctx, \{\{function.required_kwargs\}\}
    )
    assert ret
"""


def run(hub, ctx, root_directory: pathlib.Path):
    # TODO
    ...



from typing import Dict
from typing import List

from cloudspec import CloudSpecFunction


def return_type(hub, func_data: CloudSpecFunction):
    ret = ""
    if func_data.return_type:
        ret += "\n\n    Returns:\n"
        ret += f"        {func_data.return_type}"
        ret += "\n"
    return ret


def doc(hub, func_data: CloudSpecFunction):
    if func_data.doc:
        doc = func_data.doc.replace('"""', "'''")
        doc = "\n" + hub.tool.format.wrap.indent(doc, 1) + "\n"
    else:
        doc = ""
    return doc


def parse(hub, functions: Dict[str, CloudSpecFunction], targets: List[str]):
    funcs = {}
    if not functions:
        # It might be an init file
        return funcs
    for function_name in targets:
        function_data = functions[function_name]

        doc = hub.cloudspec.parse.function.doc(function_data)
        doc += hub.cloudspec.parse.param.sphinx_docs(function_data.params)
        doc += "\n\n    Returns:\n        Dict[str, Any]\n"

        param_mapping = hub.cloudspec.parse.param.mappings(function_data.params)
        funcs[function_name] = dict(
            function=dict(
                name=function_name,
                hardcoded=function_data.hardcoded,
                doc=doc,
                header_params=hub.cloudspec.parse.param.headers(function_data.params),
                required_call_params=hub.cloudspec.parse.param.callers(
                    function_data.params
                ),
            ),
            parameter=dict(
                mapping=param_mapping,
            ),
        )
    return funcs



import pathlib

from cloudspec import CloudSpec


def run(hub, ctx, root_directory: pathlib.Path or str):
    if isinstance(root_directory, str):
        root_directory = pathlib.Path(root_directory)
    cloud_spec = CloudSpec(**ctx.cloud_spec)
    states_dir = root_directory / ctx.clean_name / "states" / ctx.service_name

    for ref, plugin in cloud_spec.plugins.items():
        mod_file = hub.cloudspec.parse.plugin.touch(states_dir, ref)
        ref = hub.cloudspec.parse.plugin.ref(ctx, ref)
        state_ref = hub.cloudspec.parse.plugin.mod_ref(ctx, ref, plugin)
        to_write = hub.cloudspec.parse.plugin.header(plugin)
        if not plugin.functions:
            mod_file.write_text(to_write)
            continue
        func_data = hub.cloudspec.parse.function.parse(
            plugin.functions, targets=("present", "absent", "describe")
        )
        present_parameter = hub.cloudspec.parse.param.simple_map(
            plugin.functions["present"].params
        )
        common_params = {}
        for name, param in plugin.functions["absent"].params.items():
            if name in plugin.functions["describe"].params:
                common_params[name] = plugin.functions["describe"].params[name]

        common_params = hub.cloudspec.parse.param.mappings(common_params)

        # Create the present, absent, and describe functions; these are required for every state module
        for function_name, TEMPLATE in zip(
            ("present", "absent", "describe"),
            (
                hub.cloudspec.template.state.PRESENT_FUNCTION,
                hub.cloudspec.template.state.ABSENT_FUNCTION,
                hub.cloudspec.template.state.DESCRIBE_FUNCTION,
            ),
        ):
            template = hub.tool.jinja.template(
                f"{TEMPLATE}\n    {cloud_spec.request_format[function_name]}\n\n\n"
            )

            to_write += template.render(
                service_name=cloud_spec.service_name,
                function=dict(
                    ref=ref,
                    state_ref=f"states.{state_ref}",
                    **func_data[function_name]["function"],
                ),
                parameter=func_data[function_name]["parameter"],
                present_parameter=present_parameter,
                get_params=common_params,
            )

            mod_file.write_text(to_write)



from typing import Dict

from cloudspec import CloudSpecParam


def sphinx_docs(hub, parameters: CloudSpecParam) -> str:
    """
    Get the sphinx docs for the parameters
    """
    if not parameters:
        return ""

    ret = "\n    Args:\n"
    for param in parameters.values():
        ret += f"        {param.snaked}"
        if param.param_type:
            ret += f"({param.param_type}{', optional' if not param.required else ''})"

        ret += f": {param.doc}"
        if not ret.endswith("."):
            ret += "."
        if not param.required:
            ret += f" Defaults to {param.default}."

        if param.get("member", None):
            for nested_param_name, nested_param_data in (
                param.get("member").get("params").items()
            ):
                ret += "\n            * "
                ret += f"({nested_param_name}{', optional' if not nested_param_data.required else ''}): {nested_param_data.doc}"
        ret += "\n"

    return ret.rstrip()


def headers(hub, parameters: CloudSpecParam) -> str:
    """
    The arguments that will be put in the function definition
    """
    ret = ""

    required_params = {
        name: data for name, data in parameters.items() if data["required"]
    }
    for param_name, param_data in required_params.items():
        if param_data.get("param_type"):
            ret += f",\n    {param_data.snaked}: {param_data['param_type']}"
            if param_data.get("member", None):
                ret += nested_param(param_data.get("member"))
        else:
            ret += f",\n    {param_data.snaked}"

    unrequired_params = {
        name: data for name, data in parameters.items() if not data["required"]
    }

    for param_name, param_data in unrequired_params.items():
        param_type = param_data.get("param_type")
        if param_type:
            ret += f",\n    {param_data.snaked}: {param_type}"
            if param_data.get("member", None):
                ret += nested_param(param_data.get("member"))
        else:
            ret += f",\n    {param_data.snaked}"

        # TODO handle this case properly
        if param_data.default == "{}" or param_data.default == "[]":
            ret += f" = None"
        else:
            ret += f" = {param_data.default}"

    return ret


def nested_param(member: Dict) -> str:
    """
    Create a dataclass for a complex nested argument
    """
    ret = f'[\n        make_dataclass(\n            "{member.name}",\n            ['
    for param_name, param_data in member.get("params").items():
        ret += f"\n                "
        # TODO: Nested param name should be snaked
        ret += f'("{param_name}", {param_data.param_type}, field(default={param_data.default})), '

    # Remove last ', '
    ret = ret[0:-2]
    ret += "\n            ]\n        )\n    ]"

    return ret


def callers(hub, parameters: CloudSpecParam) -> str:
    """
    Get a mapping of the function args to the values that will be used in the final caller
    """
    ret = []

    required_params = {
        name: data for name, data in parameters.items() if data["required"]
    }
    for param_data in required_params.values():
        ret.append(f"{param_data.snaked}={param_data.default or 'value'}")

    return ", ".join(ret)


def mappings(hub, parameters: CloudSpecParam) -> Dict[str, str]:
    ret = {}
    map_params = {
        name: data
        for name, data in parameters.items()
        if data["target_type"] == "mapping"
    }
    for name, data in map_params.items():
        target = data["target"]
        if target not in ret:
            ret[target] = {}
        ret[target][name] = data.snaked

    fmt = lambda item: ", ".join(f'"{k}": {v}' for k, v in item.items())
    return {k: f"{{{fmt(v)}}}" for k, v in ret.items()}


def simple_map(hub, parameters: Dict[str, CloudSpecParam]):
    result = {}
    for k, param_data in parameters.items():
        if k == "name" and param_data.target_type == "arg":
            continue
        result[param_data.snaked] = k
    return result



from typing import Any
from typing import Dict
from typing import Tuple

import openapi3.object_base

__func_alias__ = {"type_": "type"}


def plugins(hub, ctx, paths: openapi3.object_base.Map) -> Dict[str, Any]:
    ret = {}
    for name, path in paths.items():
        assert isinstance(path, openapi3.paths.Path)
        # Get the request type that works for this request
        for request_type in path.raw_element.keys():
            func: openapi3.paths.Operation = getattr(path, request_type)
            if not func:
                continue
            subs = [hub.tool.format.case.snake(sub) for sub in func.tags]
            if not subs:
                plugin = "init"
            else:
                plugin = subs.pop()

            refs = [ctx.service_name] + subs + [plugin]
            ref = ".".join(refs)
            if ref not in ret:
                # This is the first time we have looked at this plugin
                ret[ref] = {"functions": {}, "doc": ""}
            func_name, func_data = hub.pop_create.openapi3.parse.function(name, func)
            func_data["hardcoded"] = {
                "method": request_type,
                "path": name.split(" ")[0],
            }
            ret[ref]["functions"][func_name] = func_data
    return ret


def function(
    hub,
    name: str,
    func: openapi3.paths.Operation,
) -> Tuple[str, Dict[str, Any]]:
    # This is the preferred way to get a function name
    func_name = func.operationId
    # Fallback function name based on the pets example
    if not func_name and " " in name:
        func_name = "_".join(name.split(" ")[1:]).lower()
    if not func_name and func.extensions:
        func_name = func.extensions[sorted(func.extensions.keys())[0]]
    else:
        func_name = func.summary

    # Maybe we need more fallbacks, you tell me
    if not func_name:
        # Maybe a fallback based on the path and method?
        raise AttributeError(f"Not sure how to find func name for {name}, help me out")

    func_name = hub.tool.format.case.snake(func_name)

    func_spec = {
        "doc": (func.description or "").strip(),
        "params": {
            p.name: hub.pop_create.openapi3.parse.parameter(p) for p in func.parameters
        },
    }
    return func_name, func_spec


def parameter(hub, parameter: openapi3.paths.Parameter):
    if parameter.in_ == "query":
        target_type = "mapping"
    elif parameter.in_ == "path":
        target_type = "mapping"
    elif parameter.in_ == "header":
        target_type = "mapping"
    elif parameter.in_ == "cookie":
        target_type = "mapping"
    else:
        raise ValueError(f"Unknown parameter type: {parameter.in_}")

    return {
        "required": parameter.required,
        "target_type": target_type,
        "target": parameter.in_,
        "param_type": hub.pop_create.openapi3.parse.type(parameter.schema.type),
        "doc": parameter.description or parameter.name,
    }


def type_(hub, param_type: str) -> str:
    if "integer" == param_type:
        return "int"
    elif "boolean" == param_type:
        return "bool"
    elif "number" == param_type:
        return "float"
    elif "string" == param_type:
        return "str"
    else:
        return ""



import pathlib

import pop.hub
from dict_tools.data import NamespaceDict

if __name__ == "__main__":
    root_directory = pathlib.Path.cwd()
    hub = pop.hub.Hub()
    hub.pop.sub.add(dyne_name="tool")
    hub.pop.sub.load_subdirs(hub.tool, recurse=True)
    ctx = NamespaceDict({{cookiecutter}})

    # Sanitize based on other arguments
    if ctx.has_acct_plugin:
        hub.tool.path.rmtree(root_directory / ctx.clean_name / "acct")
        hub.tool.path.rmtree(root_directory / ctx.clean_name / "tool")
        hub.tool.path.delete(
            root_directory / ctx.clean_name / "exec" / ctx.service_name / "sample.py"
        )
        hub.tool.path.delete(
            root_directory / ctx.clean_name / "states" / ctx.service_name / "sample.py"
        )



HTTP_REQUEST = """
return await hub.tool.http.session.request(
    ctx,
    method="{{{{ function.hardcoded.method }}}}",
    path=ctx.acct.endpoint_url + "{{{{ function.hardcoded.path }}}}".format(
        **{{{{ parameter.mapping.path|default({{}}) }}}}
    ),
    query_params={{{{ parameter.mapping.query|default({{}}) }}}},
    data={{{{ parameter.mapping.header|default({{}}) }}}}
)
"""



import pathlib

from cloudspec import CloudSpec


def run(hub, ctx, root_directory: pathlib.Path or str):
    if isinstance(root_directory, str):
        root_directory = pathlib.Path(root_directory)
    cloud_spec = CloudSpec(**ctx.cloud_spec)
    exec_dir = root_directory / ctx.clean_name / "exec" / ctx.service_name

    for ref, plugin in cloud_spec.plugins.items():
        mod_file = hub.cloudspec.parse.plugin.touch(exec_dir, ref)
        ref = hub.cloudspec.parse.plugin.ref(ctx, ref)
        exec_ref = hub.cloudspec.parse.plugin.mod_ref(ctx, ref, plugin)

        # Set up the base template
        if not plugin.functions:
            to_write = hub.cloudspec.parse.plugin.header(plugin)

        else:
            to_write = hub.cloudspec.parse.plugin.header(plugin)
            mod_file.write_text(to_write)
            func_data = hub.cloudspec.parse.function.parse(
                plugin.functions, targets=("get", "list", "create", "update", "delete")
            )
            present_parameter = hub.cloudspec.parse.param.simple_map(
                plugin.functions["create"].params
            )

            # Make the get, list, create, delete, and update functions; these are required for every auto_state exec module
            for function_name in "get", "list", "create", "delete", "update":
                template_str = f"{hub.cloudspec.template.exec.FUNCTION}\n    {cloud_spec.request_format[function_name]}\n\n\n"
                template = hub.tool.jinja.template(template_str)

                to_write += template.render(
                    service_name=cloud_spec.service_name,
                    function=dict(
                        ref=ref,
                        exec_ref=f"exec.{exec_ref}",
                        **func_data[function_name]["function"],
                    ),
                    parameter=func_data[function_name]["parameter"],
                    present_parameter=present_parameter,
                )

        mod_file.write_text(to_write)



import json
import pathlib

from dict_tools.data import NamespaceDict


def __init__(hub):
    hub.pop.sub.add(dyne_name="cloudspec")


def context(hub, ctx, directory: pathlib.Path):
    # If an acct plugin was passed in then we don't need to create auth plugins
    if ctx.get("simple_service_name"):
        ctx.service_name = ctx.simple_service_name
    elif not ctx.get("service_name"):
        ctx.service_name = (
            ctx.clean_name.replace("idem", "").replace("cloud", "").strip("_")
        )

    ctx.clean_api_version = ctx.get("clean_api_version")
    ctx.cloud_api_version = ctx.get("cloud_api_version")
    ctx.servers = ctx.get("servers", [""])

    if ctx.specification:
        with open(ctx.specification, "w+") as fh:
            data = json.load(fh)

        ctx.cloud_spec = data
    else:
        ctx.cloud_spec = NamespaceDict(
            api_version=ctx.get("cloud_api_version"),
            project_name=ctx.project_name,
            service_name=ctx.service_name,
            request_format=None,
            plugins={},
        )

    ctx.has_acct_plugin = bool(ctx.acct_plugin)
    if not ctx.has_acct_plugin:
        # Create auth plugins
        ctx.acct_plugin = ctx.service_name

    return ctx



import pathlib

import requests


def context(hub, ctx, directory: pathlib.Path):
    ctx = hub.pop_create.idem_cloud.init.context(ctx, directory)

    swagger = hub.pop_create.openapi3.init.read(source=hub.OPT.pop_create.specification)

    # Convert swagger to openapi3
    response = requests.post(
        "https://converter.swagger.io/api/convert",
        json=dict(swagger),
        headers={"Content-Type": "application/json"},
    )
    ctx.specification = response.json()

    # get ctx from openapi3 with changes that turn swagger into openapi3
    hub.pop_create.init.run(directory=directory, subparsers=["openapi3"], **ctx)

    return ctx
